from typing import Any
from custom_gtg import geographical_threshold_graph_custom
import networkx as nx
import numpy as np
import matplotlib.pyplot as plt
from IPython.display import display
import os


# ---------------------- #
#        Classes         #
# ---------------------- #


class BaseGraph:
    def __init__(
        self,
        n_nodes,
        upd_w=np.inf,
        upd_graph=np.inf,
        upd_plot=np.inf,
        figpath=None,
        plotable=False,
    ):
        self.n_nodes = n_nodes
        self.upd_w = upd_w
        self.upd_graph = upd_graph
        self.upd_plot = upd_plot

        if plotable:
            self.fig, self.ax = plt.subplots(dpi=150)
            self.ax.set_xlim([-0.03, 1.03])
            self.ax.set_ylim([-0.03, 1.03])
            plt.close(self.fig)

        if figpath is not None:
            self.fighpath = figpath
            self.temppath = os.path.join(self.figpath, "temp")

    def plot_snapshot(self, *args, **kwargs):
        pass

    def update_weights(self, *args, **kwargs):
        pass

    def update_graph(self, *args, **kwargs):
        pass

    def get_opponents_cpu(self, *args, **kwargs):
        pass


class GTG(BaseGraph):
    def __init__(
        self, n_nodes, theta, join="add", w0=None, posi=None, seed=None, **kwargs
    ):
        super().__init__(n_nodes, **kwargs)
        self.theta = theta
        self.posi = posi
        self.join = join
        self.seed = seed

        self.G = geographical_threshold_graph_custom(
            n_nodes,
            theta,
            dim=2,
            pos=posi,
            weight=w0,
            metric=None,
            p_dist=None,
            seed=seed,
            join=join,
        )

        if self.posi is None:
            self.posi = [self.G.nodes[i]["pos"] for i in range(n_nodes)]

        self.w = dict(enumerate(self.G.nodes[i]["weight"] for i in range(self.n_nodes)))

    def __call__(self, *args, **kwds):
        return self.G

    # Modificar la funcion para que o bien guarde en temp o bien muestre en pantalla
    def plot_snapshot(self, w_min, mcs=None, mode="save", *args, **kwargs):
        # Esto se tiene que escribir mejor
        a = np.array(list(self.w.values()))

        dead_nodes = [node for node, weight in self.w.items() if weight < w_min]

        node_size = 25 * 200 * (a)
        node_colors = ["r" if n in dead_nodes else "royalblue" for n in self.G.nodes]
        edge_colors = [
            "r" if (e[0] in dead_nodes or e[1] in dead_nodes) else "black"
            for e in self.G.edges
        ]

        self.ax.clear()
        if mcs is not None:
            self.ax.set_title(f"t = {mcs}")
        nx.draw(
            self.G,
            node_size=node_size,
            width=0.2,
            pos=self.posi,
            node_color=node_colors,
            edge_color=edge_colors,
            ax=self.ax,
        )

        if mode == "save":
            filename = os.path.join(self.temppath, "test_{:05d}.png".format(mcs))
            self.fig.savefig(filename, format="PNG")

        if mode == "show":
            display(self.fig)

    def update_weights(self, wealths, *args, **kwargs):
        self.w = dict(enumerate(wealths))
        nx.set_node_attributes(self.G, self.w, "weight")

    def update_graph(self, *args, **kwargs):
        self.G = geographical_threshold_graph_custom(
            self.n_nodes,
            self.theta,
            dim=2,
            pos=self.posi,
            weight=self.w,
            metric=None,
            p_dist=None,
            seed=self.seed,
            join=self.join,
        )

    def get_opponents_cpu(self, *args, **kwargs):
        opponents = np.full(self.n_nodes, fill_value=-1)
        for i in range(self.n_nodes):
            if neighbors := list(nx.all_neighbors(self.G, i)):
                opponents[i] = np.random.choice(neighbors)

        return opponents

    # Function inspired by cuTradeNet's toLL() function in:
    # https://github.com/Qsanti/cuTradeNet/blob/e20f29e65bcac448d7fcd17fac45746f90e8538e/cuTradeNet/Models/Utils/GraphManager.py
    def get_neigbhors_array_gpu(self, *args, **kwargs):
        """ """
        neighs = [(list(self.G.neighbors(i))) for i in range(self.n_nodes)]
        cum_neighs = np.cumsum(np.array([len(x) for x in neighs], dtype=np.int32))
        neighs = np.hstack(neighs).astype(np.int32)
        return cum_neighs, neighs


# Acá iría la clase del multigraph


# ---------------------- #
#       Funciones        #
# ---------------------- #



from numba import cuda


@cuda.jit(device=True)
def double_lock(mutex, i, j):
    first, second = (i, j) if i < j else (j, i)

    while cuda.atomic.cas(mutex, first, 0, 1) != 0:
        pass
    while cuda.atomic.cas(mutex, second, 0, 1) != 0:
        pass

    # cuda.threadfence()


@cuda.jit(device=True)
def double_unlock(mutex, i, j):
    # cuda.threadfence()
    cuda.atomic.exch(mutex, j, 0)
    cuda.atomic.exch(mutex, i, 0)



import numpy as np
import cupy as cp


class BaseModel:
    def __init__(
        self,
        n_agents,
        w_min=1e-17,
        G=None,
        save_every=np.inf,
        upd_w_every=np.inf,
        upd_graph_every=np.inf,
        plot_every=np.inf,
    ):  
        self.n_agents = n_agents
        self.w_min = w_min
        self.G = G
        if self.G is not None:
            self.update_w = upd_w_every
            self.update_links = upd_graph_every
            self.plot = plot_every
        self.every = save_every


class GPUModel(BaseModel):
    def __init__(self, w_0, **kwargs):  # sourcery skip: assign-if-exp
        super().__init__(**kwargs)
        # Initialize n agents with random risks and wealth between (0, 1]
        # and normalize wealth
        # n[i, 0] is the wealth and n[i, 1] is the risk of agent i
        self.w = cp.random.rand(self.n_agents).astype(cp.float32)
        self.r = cp.random.rand(self.n_agents).astype(cp.float32)
        if w_0 is not None:
            self.w = w_0
        else:
            self.w = self.w / (np.sum(self.w))
        # self.gini = [self.get_gini()]
        # self.n_active = [self.get_actives()]

    def MCS(self, steps):
        """
        Main MC loop
        """
        for mcs in range(steps):
            j = self.get_opponents()

            for i, ji in enumerate(j):
                # Check both agents have w > w_min and node is not isolated
                if self.is_valid(i, ji) and ji != -1:
                    dw = self.get_dw(i, ji)
                    winner = self.choose_winner(i, ji)
                    dw = np.where(winner == i, dw, -dw)
                    self.update_wealth(i, ji, dw)

            
            # After self.every MCS append new Gini index
            # if (mcs + 1) % self.every == 0:
                # self.gini.append(self.get_gini())
                # self.n_active.append(self.get_actives())

class GPUEnsemble:
    def __init__(self):
        


class CPUModel(BaseModel):
    def __init__(self, w_0, **kwargs):  # sourcery skip: assign-if-exp
        super().__init__(**kwargs)
        # Initialize n agents with random risks and wealth between (0, 1]
        # and normalize wealth
        # n[i, 0] is the wealth and n[i, 1] is the risk of agent i
        self.w = cp.random.rand(self.n_agents).astype(cp.float32)
        self.r = cp.random.rand(self.n_agents).astype(cp.float32)
        if w_0 is not None:
            self.w = w_0
        else:
            self.w = self.w / (np.sum(self.w))
        # self.gini = [self.get_gini()]
        # self.n_active = [self.get_actives()]

    def get_opponents(self):
        if self.G is None:
            random_array = np.random.randint(0, self.N, self.N)
            indices = np.arange(0, self.N)
            # Create array of random numbers that are not equal to the index
            # If i=j then assign j'=i+1 (j'=0 if i=N-1)
            random_array = np.where(
                random_array == indices, (random_array + 1) % self.N, random_array
            )
        else:
            random_array = np.full(self.N, fill_value=-1)
            for i in range(self.N):
                if neighbors := list(nx.all_neighbors(self.G, i)):
                    random_array[i] = np.random.choice(neighbors)

        return random_array

    def is_valid(self, i, j):
        # Check if both agents have w > w_min
        return (self.n[i, 0] > self.w_min) and (self.n[j, 0] > self.w_min)

    def get_dw(self, i, j):
        return np.minimum(self.n[i, 0] * self.n[i, 1], self.n[j, 0] * self.n[j, 1])

    def get_gini(self):
        w = np.sort(self.n[:, 0])
        p_cumsum = np.cumsum(w) / np.sum(w)
        B = np.sum(p_cumsum) / self.N
        return 1 + 1 / self.N - 2 * B

    def get_actives(self):
        return np.sum(self.n[:, 0] > self.w_min)

    def get_liquidity():
        return

    def update_wealth(self, i, j, dw):
        self.n[i, 0] += dw
        self.n[j, 0] -= dw

    def choose_winner(self, i, j):
        raise Exception("You need to choose a valid model.")

    def update_weights(self):
        w = dict(enumerate(self.n[:, 0]))
        nx.set_node_attributes(self.G, w, "weight")

    def plot_snapshot(self, mcs):
        ## Esto se tiene que borrar y escribir mejor
        w = dict(enumerate(self.n[:, 0]))
        a = np.array(list(w.values()))

        dead_nodes = [node for node, weight in w.items() if weight < self.w_min]

        node_size = 500 * np.sqrt(a)
        node_colors = plt.cm.coolwarm(100 * a)
        edge_colors = [
            "r" if (e[0] in dead_nodes or e[1] in dead_nodes) else "black"
            for e in self.G.edges
        ]

        filename = os.path.join(self.temppath, "test_{:05d}.png".format(mcs))

        self.ax.clear()
        self.ax.set_title(f"t = {mcs}")
        nx.draw(
            self.G,
            node_size=node_size,
            width=0.2,
            pos=self.posi,
            node_color=node_colors,
            edge_color=edge_colors,
            ax=self.ax,
        )
        self.fig.savefig(filename, format="PNG")

    def MCS(self, steps):
        """
        Main MC loop
        """
        for mcs in range(steps):
            if mcs % self.plot == 0:
                self.plot_snapshot(mcs)

            j = self.get_opponents()

            for i, ji in enumerate(j):
                # Check both agents have w > w_min and node is not isolated
                if self.is_valid(i, ji) and ji != -1:
                    dw = self.get_dw(i, ji)
                    winner = self.choose_winner(i, ji)
                    dw = np.where(winner == i, dw, -dw)
                    self.update_wealth(i, ji, dw)

            # After self.update_w update weights
            if mcs % self.update_w == 0:
                self.update_weights()

            # Recompute the links if the network is dynamic
            if (mcs + 1) % self.update_links == 0 and self.G is not None:
                self.G = nx.geographical_threshold_graph(
                    self.N,
                    theta=self.theta,
                    weight=self.n[:, 0],
                    dim=2,
                    pos=self.posi,
                    additive=self.additive,
                )
            # After self.every MCS append new Gini index
            if (mcs + 1) % self.every == 0:
                self.gini.append(self.get_gini())
                self.n_active.append(self.get_actives())



from locks import *
from numba import cuda
from numba.cuda.random import xoroshiro128p_uniform_float32



# Kernel de Yard-Sale
@cuda.jit
def k_ys_mcs(n_agents, w, r, mutex, cum_neighs, neighs, wmin, f, rng_states, mcs):
    tid = cuda.grid(1)

    if tid >= n_agents:
        return

    cuda.syncthreads()

    if tid == 0:
        n_neighs = cum_neighs[tid]

    else:
        n_neighs = cum_neighs[tid] - cum_neighs[tid - 1]

    for _ in range(mcs):
        if n_neighs[tid] != 0:
            # Choose random index between 0 and n_neighs - 1
            rand_neigh = int(
                xoroshiro128p_uniform_float32(rng_states, tid) * (n_neighs)
            )
            # Assigns tid's opponent to his rand_neigh'th neighbor
            opp = neighs[cum_neighs[tid] + rand_neigh]


            if w[tid] > wmin and w[opp] > wmin:
                double_lock(mutex, tid, opp)

                dw = min(r[tid] * w[tid], r[opp] * w[opp])

                # Compute the probability of winning
                p = 0.5 + f * abs(w[tid] - w[opp]) / (w[tid] + w[opp])
                rand = xoroshiro128p_uniform_float32(rng_states, tid)
                # Determine the winner and loser
                dw = dw if rand <= p else -dw
                w[tid] += dw
                w[opp] -= dw

                double_unlock(mutex, tid, opp)

        cuda.syncthreads()
